import functools
from flask import abort, request
from authserv import app
from authserv import protocol


class RateLimit(object):

    prefix = 'authserv/'

    def __init__(self, count, period):
        self.count = count
        self.period = period

    def check(self, mc, key):
        key = self.prefix + key
        result = mc.incr(key)
        if result is None:
            result = 1
            if not mc.add(key, result, time=self.period):
                result = mc.incr(key)
                if result is None:
                    # Memcache is failing.
                    return True
        return result <= self.count


def ratelimit(header=None, param=None, count=0, period=0):
    if not header and not param:
        raise Exception('Must set either header or param')
    rl = RateLimit(count, period)
    required_parts = 2
    if header:
        required_parts += 1
    if param:
        required_parts += 1

    def decoratorfn(fn):
        @functools.wraps(fn)
        def _ratelimit(*args, **kwargs):
            parts = [request.method, request.path]
            if header:
                value = request.environ.get(header)
                if value:
                    parts.append('%s=%s' % (header, value))
            if param:
                value = request.form.get(param)
                if value:
                    parts.append('%s=%s' % (param, value))
            if len(parts) >= required_parts:
                key = ' '.join(parts)
                if not rl.check(app.memcache, key):
                    app.logger.debug('ratelimited: %s', key)
                    abort(503)
            return fn(*args, **kwargs)
        return _ratelimit
    return decoratorfn


class BlackList(object):

    prefix = 'bl/'

    def __init__(self, count, period, ttl):
        self.rl = RateLimit(count, period)
        self.ttl = ttl

    def check(self, mc, key):
        key = self.prefix + key
        result = mc.get(key)
        return result is None

    def incr(self, mc, key):
        key = self.prefix + key
        if not self.rl.check(mc, key):
            mc.add(key, 1, time=self.ttl)


def blacklist(args_idx=None, count=0, period=0, ttl=0):
    if not args_idx:
        raise Exception('Must set args_idx')
    bl = BlackList(count, period, ttl)

    def decoratorfn(fn):
        @functools.wraps(fn)
        def _blacklist(*args, **kwargs):
            parts = [args[x] for x in args_idx if args[x]]
            if not parts:
                return fn(*args, **kwargs)
            key = ' '.join(parts)
            if not bl.check(app.memcache, key):
                app.logger.debug('blacklisted %s', key)
                return protocol.ERR_AUTHENTICATION_FAILURE
            result = fn(*args, **kwargs)
            if result != protocol.OK:
                bl.incr(app.memcache, key)
            return result
        return _blacklist
    return decoratorfn

